# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from functools import wraps

import pytest

from snowflake.hypothesis_snowpark.logging_utils import log


def test_log_decorator_logs_function_call_with_args(caplog: pytest.LogCaptureFixture):
    """Test that the log decorator logs the function call with the arguments."""

    @log
    def sample_function(a: int, b: int = 2):
        return a + b

    with caplog.at_level(logging.DEBUG):
        result = sample_function(1, b=3)

    assert result == 4
    assert "sample_function called with args 1, b=3" in caplog.text


def test_log_decorator_logs_function_call_without_args(
    caplog: pytest.LogCaptureFixture,
):
    """Test that the log decorator logs the function call without the arguments."""

    @log
    def sample_function():
        return 1

    with caplog.at_level(logging.DEBUG):
        result = sample_function()

    assert result == 1
    assert "sample_function called with args" in caplog.text


def test_log_decorator_logs_exceptions(caplog: pytest.LogCaptureFixture):
    """Test that the log decorator logs exceptions."""
    exception_msg = "An error occurred"

    @log
    def sample_function():
        raise ValueError(exception_msg)

    with pytest.raises(ValueError, match=exception_msg), caplog.at_level(logging.DEBUG):
        sample_function()

    assert "An error occurred in sample_function" in caplog.text
    assert exception_msg in caplog.text


def test_log_decorator_custom_logger(caplog: pytest.LogCaptureFixture):
    """Test that the log decorator logs to a custom logger."""
    logger_name = "custom_logger"
    custom_logger = logging.getLogger(logger_name)

    @log(logger=custom_logger)
    def sample_function(a: int, b: int):
        return a + b

    with caplog.at_level(logging.DEBUG, logger=logger_name):
        result = sample_function(1, 2)

    assert result == 3
    assert "sample_function called with args 1, 2" in caplog.text


def test_log_decorator_no_log_args(caplog: pytest.LogCaptureFixture):
    """Test that the log decorator does not log the arguments when log_args is False."""

    @log(log_args=False)
    def sample_function(a: int, b: int):
        return a + b

    with caplog.at_level(logging.DEBUG):
        result = sample_function(1, 2)

    assert result == 3
    assert "sample_function called with args" not in caplog.text


def test_log_decorator_with_other_decorators(caplog: pytest.LogCaptureFixture):
    """Test that the log decorator works with other decorators."""

    def dummy_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            raise ValueError("dummy_decorator called")

        return wrapper

    @dummy_decorator
    @log
    def sample_function_1():
        return 1

    @log
    @dummy_decorator
    def sample_function_2():
        return 2

    with caplog.at_level(logging.DEBUG), pytest.raises(
        ValueError, match="dummy_decorator called"
    ):
        sample_function_1()

    assert "An error occurred in sample_function" not in caplog.text

    with caplog.at_level(logging.DEBUG), pytest.raises(
        ValueError, match="dummy_decorator called"
    ):
        sample_function_2()

    assert "An error occurred in sample_function" in caplog.text



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import inspect
from pathlib import Path
import tempfile

from unittest.mock import patch

import hypothesis.strategies as st
import pandera as pa
import pytest

from hypothesis import HealthCheck, given, settings

from snowflake.hypothesis_snowpark import dataframe_strategy
from snowflake.snowpark import DataFrame, Session
from snowflake.snowpark.types import (
    ArrayType,
    LongType,
    StructField,
    StructType,
)
from snowflake.hypothesis_snowpark.io_utils.io_file_manager import IODefaultStrategy
from snowflake.hypothesis_snowpark.io_utils.io_file_manager import (
    get_io_file_manager,
)
from .telemetry_compare_utils import (
    validate_telemetry_file_output,
    reset_telemetry_util,
)

SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME = "snowpark-checkpoints-output"

TELEMETRY_FOLDER = "telemetry"


@pytest.fixture(scope="function")
def telemetry_output_path():
    folder = os.urandom(8).hex()
    directory = Path(tempfile.gettempdir()).resolve() / folder
    os.makedirs(directory)
    telemetry_dir = directory / TELEMETRY_FOLDER
    return telemetry_dir


@given(data=st.data())
@settings(deadline=None, max_examples=1, suppress_health_check=list(HealthCheck))
def test_io_strategy_from_json(
    data: st.DataObject, session: Session, telemetry_output_path: Path
):
    try:
        # Arrange
        class TestStrategy(IODefaultStrategy):
            pass

        number_of_methods = inspect.getmembers(
            IODefaultStrategy, predicate=inspect.isfunction
        )
        strategy = TestStrategy()
        get_io_file_manager().set_strategy(strategy)

        with (
            patch.object(strategy, "getcwd", wraps=strategy.getcwd) as getcwd_spy,
            patch.object(strategy, "ls", wraps=strategy.ls) as ls_spy,
            patch.object(strategy, "mkdir", wraps=strategy.mkdir) as mkdir_spy,
            patch.object(strategy, "write", wraps=strategy.write) as write_spy,
            patch.object(strategy, "read", wraps=strategy.read) as read_spy,
            patch.object(
                strategy, "read_bytes", wraps=strategy.read_bytes
            ) as read_bytes_spy,
            patch.object(
                strategy, "file_exists", wraps=strategy.file_exists
            ) as file_exists_spy,
            patch.object(
                strategy, "folder_exists", wraps=strategy.folder_exists
            ) as folder_exists_spy,
        ):
            telemetry_manager = reset_telemetry_util()
            telemetry_manager.set_sc_output_path(telemetry_output_path)
            # Act
            strategy = dataframe_strategy(
                schema="test/resources/test_io.json",
                session=session,
            )
            df = data.draw(strategy)

            expected_schema = StructType(
                [
                    StructField("array_column", ArrayType(), True),
                ]
            )

            # Assert
            assert df.schema == expected_schema
            assert isinstance(df, DataFrame)
            assert len(number_of_methods) == 9
            read_bytes_spy.assert_not_called()
            folder_exists_spy.assert_not_called()

            assert file_exists_spy.call_count == 1
            assert read_spy.call_count == 3
            assert write_spy.call_count == 1
            assert mkdir_spy.call_count == 3
            assert getcwd_spy.call_count == 1
            assert ls_spy.call_count == 1
            validate_telemetry_file_output(
                "test_io_strategy_from_json_telemetry.json", str(telemetry_output_path)
            )
    finally:
        get_io_file_manager().set_strategy(IODefaultStrategy())


@given(data=st.data())
@settings(deadline=None, max_examples=1, suppress_health_check=list(HealthCheck))
def test_io_strategy_from_object(
    data: st.DataObject, session: Session, telemetry_output_path: Path
):
    try:
        # Arrange
        class TestStrategy(IODefaultStrategy):
            pass

        number_of_methods = inspect.getmembers(
            IODefaultStrategy, predicate=inspect.isfunction
        )
        strategy = TestStrategy()
        get_io_file_manager().set_strategy(strategy)

        with (
            patch.object(strategy, "getcwd", wraps=strategy.getcwd) as getcwd_spy,
            patch.object(strategy, "ls", wraps=strategy.ls) as ls_spy,
            patch.object(strategy, "mkdir", wraps=strategy.mkdir) as mkdir_spy,
            patch.object(strategy, "write", wraps=strategy.write) as write_spy,
            patch.object(strategy, "read", wraps=strategy.read) as read_spy,
            patch.object(
                strategy, "read_bytes", wraps=strategy.read_bytes
            ) as read_bytes_spy,
            patch.object(
                strategy, "file_exists", wraps=strategy.file_exists
            ) as file_exists_spy,
            patch.object(
                strategy, "folder_exists", wraps=strategy.folder_exists
            ) as folder_exists_spy,
        ):
            telemetry_manager = reset_telemetry_util()
            telemetry_manager.set_sc_output_path(telemetry_output_path)
            schema = pa.DataFrameSchema(
                {"short_column": pa.Column(pa.Int16, nullable=True)}
            )

            # Act
            strategy = dataframe_strategy(
                schema=schema,
                session=session,
            )
            df = data.draw(strategy)

            expected_schema = StructType(
                [StructField("short_column", LongType(), True)]
            )

            # Assert
            assert df.schema == expected_schema
            assert isinstance(df, DataFrame)
            assert len(number_of_methods) == 9
            file_exists_spy.assert_not_called()
            read_bytes_spy.assert_not_called()
            folder_exists_spy.assert_not_called()

            assert read_spy.call_count == 1
            assert write_spy.call_count == 1
            assert mkdir_spy.call_count == 3
            assert getcwd_spy.call_count == 1
            assert ls_spy.call_count == 1
            validate_telemetry_file_output(
                "test_io_strategy_from_object_telemetry.json",
                str(telemetry_output_path),
            )
    finally:
        get_io_file_manager().set_strategy(IODefaultStrategy())



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime as dt

from decimal import Decimal
from typing import Final

import hypothesis.strategies as st
import pandas as pd
import pytest

from hypothesis import HealthCheck, given, settings

from snowflake.hypothesis_snowpark.constants import (
    CUSTOM_DATA_FORMAT_KEY,
    CUSTOM_DATA_MAX_KEY,
    CUSTOM_DATA_MAX_SIZE_KEY,
    CUSTOM_DATA_MIN_KEY,
    CUSTOM_DATA_MIN_SIZE_KEY,
    CUSTOM_DATA_NAME_KEY,
    CUSTOM_DATA_TYPE_KEY,
    CUSTOM_DATA_VALUE_TYPE_KEY,
    PYSPARK_ARRAY_TYPE,
    PYSPARK_BINARY_TYPE,
    PYSPARK_DATE_TYPE,
    PYSPARK_INTEGER_TYPE,
    PYSPARK_STRING_TYPE,
)
from snowflake.hypothesis_snowpark.custom_strategies import (
    array_strategy,
    binary_strategy,
    boolean_strategy,
    byte_strategy,
    date_strategy,
    decimal_strategy,
    float_strategy,
    integer_strategy,
    long_strategy,
    short_strategy,
    string_strategy,
    update_pandas_df_strategy,
)
from snowflake.hypothesis_snowpark.strategy_register import register_strategy


MAX_EXAMPLES: Final[int] = 5


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_array_strategy(data: st.DataObject):
    min_size = 1
    max_size = 5
    result = data.draw(
        array_strategy(dtype=PYSPARK_INTEGER_TYPE, min_size=min_size, max_size=max_size)
    )
    assert isinstance(result, list)
    assert min_size <= len(result) <= max_size
    assert all(isinstance(value, int) for value in result)


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES, suppress_health_check=list(HealthCheck))
def test_array_strategy_default_values(data: st.DataObject):
    result = data.draw(array_strategy(dtype=PYSPARK_STRING_TYPE))
    assert isinstance(result, list)
    assert len(result) >= 0
    assert all(isinstance(value, str) for value in result)


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_array_strategy_unsupported_dtype(data: st.DataObject):
    dtype = "unsupported_dtype"
    with pytest.raises(
        ValueError, match=f"Not implemented SearchStrategy for arrays of type '{dtype}'"
    ):
        data.draw(array_strategy(dtype=dtype))


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_binary_strategy(data: st.DataObject):
    min_size = 1
    max_size = 5
    result = data.draw(binary_strategy(min_size=min_size, max_size=max_size))
    assert isinstance(result, bytes)
    assert min_size <= len(result) <= max_size


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_binary_strategy_default_values(data: st.DataObject):
    result = data.draw(binary_strategy())
    assert isinstance(result, bytes)
    assert len(result) >= 0


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_boolean_strategy(data: st.DataObject):
    result = data.draw(boolean_strategy())
    assert isinstance(result, bool)


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_byte_strategy(data: st.DataObject):
    result = data.draw(byte_strategy())
    assert isinstance(result, int)
    assert -128 <= result <= 127


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_date_strategy(data: st.DataObject):
    min_date = dt.date(2021, 1, 1)
    max_date = dt.date(2021, 12, 31)
    result = data.draw(date_strategy(min_value=min_date, max_value=max_date))
    assert isinstance(result, dt.date)
    assert min_date <= result <= max_date


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_date_strategy_default_values(data: st.DataObject):
    result = data.draw(date_strategy())
    assert isinstance(result, dt.date)
    assert dt.date.min <= result <= dt.date.max


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_decimal_strategy(data: st.DataObject):
    min_value = 0
    max_value = 10
    result = data.draw(decimal_strategy(min_value=min_value, max_value=max_value))
    assert isinstance(result, Decimal)
    assert min_value <= result <= max_value


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_decimal_strategy_default_values(data: st.DataObject):
    result = data.draw(decimal_strategy())
    assert isinstance(result, Decimal)


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_float_strategy(data: st.DataObject):
    min_value = 0
    max_value = 10
    result = data.draw(float_strategy(min_value=min_value, max_value=max_value))
    assert isinstance(result, float)
    assert min_value <= result <= max_value


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_float_strategy_default_values(data: st.DataObject):
    result = data.draw(float_strategy())
    assert isinstance(result, float)


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_integer_strategy(data: st.DataObject):
    min_value = 0
    max_value = 10
    result = data.draw(integer_strategy(min_value=min_value, max_value=max_value))
    assert isinstance(result, int)
    assert min_value <= result <= max_value


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_integer_strategy_default_values(data: st.DataObject):
    result = data.draw(integer_strategy())
    assert isinstance(result, int)


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_long_strategy(data: st.DataObject):
    result = data.draw(long_strategy())
    assert isinstance(result, int)
    assert -(2**63) <= result <= (2**63) - 1


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_short_strategy(data: st.DataObject):
    result = data.draw(short_strategy())
    assert isinstance(result, int)
    assert -(2**15) <= result <= 2**15 - 1


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_string_strategy(data: st.DataObject):
    min_size = 1
    max_size = 5
    result = data.draw(string_strategy(min_size=min_size, max_size=max_size))
    assert isinstance(result, str)
    assert min_size <= len(result) <= max_size


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_string_strategy_default_values(data: st.DataObject):
    result = data.draw(string_strategy())
    assert isinstance(result, str)
    assert len(result) >= 0


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_update_pandas_df_strategy_empty_columns(data: st.DataObject):
    df = pd.DataFrame({"col1": [1, 2, 3]})
    result_df = data.draw(update_pandas_df_strategy(df, []))
    pd.testing.assert_frame_equal(df, result_df)


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_update_pandas_df_strategy_empty_df(data: st.DataObject):
    col_name = "col1"
    dummy_dtype = "dummy_dtype"
    register_strategy(dummy_dtype)(st.integers)

    df = pd.DataFrame(columns=[col_name])
    columns = [{CUSTOM_DATA_TYPE_KEY: dummy_dtype, CUSTOM_DATA_NAME_KEY: col_name}]
    result_df = data.draw(update_pandas_df_strategy(df, columns))

    assert len(result_df) == 0


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_update_pandas_df_strategy_array_type(data: st.DataObject):
    col_name = "col1"
    min_size = 1
    max_size = 3

    df = pd.DataFrame({col_name: ["a", "b", "c"]})
    columns = [
        {
            CUSTOM_DATA_TYPE_KEY: PYSPARK_ARRAY_TYPE,
            CUSTOM_DATA_VALUE_TYPE_KEY: PYSPARK_INTEGER_TYPE,
            CUSTOM_DATA_MIN_SIZE_KEY: min_size,
            CUSTOM_DATA_MAX_SIZE_KEY: max_size,
            CUSTOM_DATA_NAME_KEY: col_name,
        }
    ]
    result_df = data.draw(update_pandas_df_strategy(df, columns))

    assert result_df[col_name].apply(lambda x: isinstance(x, list)).all()
    assert result_df[col_name].apply(lambda x: min_size <= len(x) <= max_size).all()
    assert result_df[col_name].apply(lambda x: all(isinstance(i, int) for i in x)).all()


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_update_pandas_df_strategy_binary_type(data: st.DataObject):
    col_name = "col1"
    min_size = 1
    max_size = 3

    df = pd.DataFrame({col_name: ["a", "b", "c"]})
    columns = [
        {
            CUSTOM_DATA_TYPE_KEY: PYSPARK_BINARY_TYPE,
            CUSTOM_DATA_MIN_SIZE_KEY: min_size,
            CUSTOM_DATA_MAX_SIZE_KEY: max_size,
            CUSTOM_DATA_NAME_KEY: col_name,
        }
    ]
    result_df = data.draw(update_pandas_df_strategy(df, columns))

    assert result_df[col_name].apply(lambda x: isinstance(x, bytes)).all()
    assert result_df[col_name].apply(lambda x: min_size <= len(x) <= max_size).all()


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_update_pandas_df_strategy_unsupported_dtype(data: st.DataObject):
    col_name = "col1"
    dtype = "unsupported_dtype"

    df = pd.DataFrame({col_name: [1, 2, 3]})
    columns = [{CUSTOM_DATA_TYPE_KEY: dtype}]

    with pytest.raises(
        ValueError, match=f"Unsupported custom strategy for data type: {dtype}"
    ):
        data.draw(update_pandas_df_strategy(df, columns))


@given(data=st.data())
@settings(max_examples=MAX_EXAMPLES)
def test_update_pandas_df_strategy_date_type(data: st.DataObject):
    col_name = "col1"
    date_format = "%Y-%m-%d"
    min_date = dt.datetime.strptime("2021-01-01", date_format).date()
    max_date = dt.datetime.strptime("2021-12-31", date_format).date()

    df = pd.DataFrame({col_name: [1, 2, 3]})
    columns = [
        {
            CUSTOM_DATA_TYPE_KEY: PYSPARK_DATE_TYPE,
            CUSTOM_DATA_FORMAT_KEY: date_format,
            CUSTOM_DATA_MIN_KEY: str(min_date),
            CUSTOM_DATA_MAX_KEY: str(max_date),
            CUSTOM_DATA_NAME_KEY: col_name,
        }
    ]
    result_df = data.draw(update_pandas_df_strategy(df, columns))

    assert len(result_df) == len(df)
    assert result_df[col_name].apply(lambda x: isinstance(x, dt.date)).all()
    assert result_df[col_name].apply(lambda x: min_date <= x <= max_date).all()



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime as dt
import logging

from decimal import Decimal
from fractions import Fraction
from typing import Optional, Union

import hypothesis.strategies as st
import pandas as pd

from snowflake.hypothesis_snowpark.constants import (
    CUSTOM_DATA_FORMAT_KEY,
    CUSTOM_DATA_MAX_KEY,
    CUSTOM_DATA_MAX_SIZE_KEY,
    CUSTOM_DATA_MIN_KEY,
    CUSTOM_DATA_MIN_SIZE_KEY,
    CUSTOM_DATA_NAME_KEY,
    CUSTOM_DATA_TYPE_KEY,
    CUSTOM_DATA_VALUE_TYPE_KEY,
    PYSPARK_ARRAY_TYPE,
    PYSPARK_BINARY_TYPE,
    PYSPARK_BOOLEAN_TYPE,
    PYSPARK_BYTE_TYPE,
    PYSPARK_DATE_TYPE,
    PYSPARK_DECIMAL_TYPE,
    PYSPARK_DOUBLE_TYPE,
    PYSPARK_FLOAT_TYPE,
    PYSPARK_INTEGER_TYPE,
    PYSPARK_LONG_TYPE,
    PYSPARK_SHORT_TYPE,
    PYSPARK_STRING_TYPE,
)
from snowflake.hypothesis_snowpark.strategy_register import (
    register_strategy,
    snowpark_strategies,
)


LOGGER = logging.getLogger(__name__)


@register_strategy(PYSPARK_ARRAY_TYPE)
@st.composite
def array_strategy(
    draw: st.DrawFn, dtype: str, min_size: int = 0, max_size: Optional[int] = None
) -> list:
    """Generate a list of values for a given data type.

    Args:
        draw: The Hypothesis draw function.
        dtype: The data type of the array.
        min_size: The minimum size of the array.
        max_size: The maximum size of the array.

    Returns:
        A list of values for the given data type.

    """
    if dtype not in snowpark_strategies:
        raise ValueError(f"Not implemented SearchStrategy for arrays of type '{dtype}'")

    strategy = snowpark_strategies[dtype]
    return draw(st.lists(strategy(), min_size=min_size, max_size=max_size))


@register_strategy(PYSPARK_BINARY_TYPE)
@st.composite
def binary_strategy(
    draw: st.DrawFn, min_size: int = 0, max_size: Optional[int] = None
) -> bytes:
    """Generate a binary data (bytes) with a given length.

    Args:
        draw: The Hypothesis draw function.
        min_size: The minimum length of the binary data.
        max_size: The maximum length of the binary data.

    Returns:
        A binary data (bytes) with a given length.

    """
    return draw(st.binary(min_size=min_size, max_size=max_size))


@register_strategy(PYSPARK_BOOLEAN_TYPE)
@st.composite
def boolean_strategy(draw: st.DrawFn) -> bool:
    """Generate a boolean value.

    Args:
        draw: The Hypothesis draw function.

    Returns:
        A boolean value.

    """
    return draw(st.booleans())


@register_strategy(PYSPARK_BYTE_TYPE)
@st.composite
def byte_strategy(draw: st.DrawFn) -> int:
    """Generate a byte value.

    Args:
        draw: The Hypothesis draw function.

    Returns:
        A byte value.

    """
    return draw(st.integers(min_value=-128, max_value=127))


@register_strategy(PYSPARK_DATE_TYPE)
@st.composite
def date_strategy(
    draw: st.DrawFn, min_value: dt.date = dt.date.min, max_value: dt.date = dt.date.max
) -> dt.date:
    """Generate a date value within a given range.

    Args:
        draw: The Hypothesis draw function.
        min_value: The minimum date value.
        max_value: The maximum date value.

    Returns:
        A date value within a given range.

    """
    return draw(st.dates(min_value=min_value, max_value=max_value))


@register_strategy(PYSPARK_DECIMAL_TYPE)
@st.composite
def decimal_strategy(
    draw: st.DrawFn,
    min_value: Union[int, float, Fraction, Decimal, str, None] = None,
    max_value: Union[int, float, Fraction, Decimal, str, None] = None,
) -> Decimal:
    """Generate a decimal value within a given range.

    Args:
        draw: The Hypothesis draw function.
        min_value: The minimum value of the decimal.
        max_value: The maximum value of the decimal.

    Returns:
        A decimal value within a given range.

    """
    return draw(st.decimals(min_value=min_value, max_value=max_value))


@register_strategy(PYSPARK_FLOAT_TYPE)
@register_strategy(PYSPARK_DOUBLE_TYPE)
@st.composite
def float_strategy(
    draw: st.DrawFn,
    min_value: Union[int, float, Fraction, Decimal, None] = None,
    max_value: Union[int, float, Fraction, Decimal, None] = None,
) -> float:
    """Generate a float value within a given range.

    Args:
        draw: The Hypothesis draw function.
        min_value: The minimum value of the float.
        max_value: The maximum value of the float.

    Returns:
        A float value within a given range.

    """
    return draw(st.floats(min_value=min_value, max_value=max_value))


@register_strategy(PYSPARK_INTEGER_TYPE)
@st.composite
def integer_strategy(
    draw: st.DrawFn,
    min_value: Optional[int] = None,
    max_value: Optional[int] = None,
) -> int:
    """Generate an integer value within a given range.

    Args:
        draw: The Hypothesis draw function.
        min_value: The minimum value of the integer.
        max_value: The maximum value of the integer.

    Returns:
        An integer value within a given range.

    """
    return draw(st.integers(min_value=min_value, max_value=max_value))


@register_strategy(PYSPARK_LONG_TYPE)
@st.composite
def long_strategy(draw: st.DrawFn) -> int:
    """Generate a long value.

    Args:
        draw: The Hypothesis draw function.

    Returns:
        A long value.

    """
    return draw(st.integers(min_value=-(2**63), max_value=2**63 - 1))


@register_strategy(PYSPARK_SHORT_TYPE)
@st.composite
def short_strategy(draw: st.DrawFn) -> int:
    """Generate a short value.

    Args:
        draw: The Hypothesis draw function.

    Returns:
        A short value.

    """
    return draw(st.integers(min_value=-(2**15), max_value=2**15 - 1))


@register_strategy(PYSPARK_STRING_TYPE)
@st.composite
def string_strategy(
    draw: st.DrawFn, min_size: int = 0, max_size: Optional[int] = None
) -> str:
    """Generate a string value with a given length.

    Args:
        draw: The Hypothesis draw function.
        min_size: The minimum length of the string.
        max_size: The maximum length of the string.

    Returns:
        A string value with a given length

    """
    return draw(st.text(min_size=min_size, max_size=max_size))


@st.composite
def update_pandas_df_strategy(
    draw: st.DrawFn,
    pandas_df: pd.DataFrame,
    columns: list[dict],
):
    """Apply custom strategies to a Pandas DataFrame based on the custom data.

    Args:
        draw: The Hypothesis draw function.
        pandas_df: The Pandas DataFrame to apply custom strategies to.
        columns: The custom data columns to apply strategies to.

    Returns:
        A Pandas DataFrame with custom strategies applied.

    """
    pandas_df_copy = pandas_df.copy()
    number_of_rows = len(pandas_df_copy)

    for column in columns:
        column_dtype = column.get(CUSTOM_DATA_TYPE_KEY)
        strategy = snowpark_strategies.get(column_dtype, None)

        if strategy is None:
            raise ValueError(
                f"Unsupported custom strategy for data type: {column_dtype}"
            )

        strategy_kwargs = {}

        if column_dtype == PYSPARK_ARRAY_TYPE:
            strategy_kwargs = {
                "dtype": column.get(CUSTOM_DATA_VALUE_TYPE_KEY),
                "min_size": column.get(CUSTOM_DATA_MIN_SIZE_KEY, 0),
                "max_size": column.get(CUSTOM_DATA_MAX_SIZE_KEY, None),
            }
        elif column_dtype == PYSPARK_BINARY_TYPE:
            strategy_kwargs = {
                "min_size": column.get(CUSTOM_DATA_MIN_SIZE_KEY, 0),
                "max_size": column.get(CUSTOM_DATA_MAX_SIZE_KEY, None),
            }
        elif column_dtype == PYSPARK_DATE_TYPE:
            date_format = column.get(CUSTOM_DATA_FORMAT_KEY)
            min_value = dt.datetime.strptime(
                column.get(CUSTOM_DATA_MIN_KEY, dt.date.min), date_format
            )
            max_value = dt.datetime.strptime(
                column.get(CUSTOM_DATA_MAX_KEY, dt.date.max), date_format
            )
            strategy_kwargs = {
                "min_value": min_value.date(),
                "max_value": max_value.date(),
            }

        column_name = column.get(CUSTOM_DATA_NAME_KEY)
        pandas_df_copy[column_name] = draw(
            st.lists(
                strategy(**strategy_kwargs),
                min_size=number_of_rows,
                max_size=number_of_rows,
            )
        )
        LOGGER.debug(
            "Applied custom strategy '%s' with kwargs %s to column '%s'",
            strategy.__name__,
            strategy_kwargs,
            column_name,
        )

    return pandas_df_copy



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional


class EnvStrategy(ABC):

    """An abstract base class that defines methods for file and directory operations.

    Subclasses should implement these methods to provide environment-specific behavior.
    """

    @abstractmethod
    def mkdir(self, path: str, exist_ok: bool = False) -> None:
        """Create a directory.

        Args:
            path: The name of the directory to create.
            exist_ok: If False, an error is raised if the directory already exists.

        """

    @abstractmethod
    def folder_exists(self, path: str) -> bool:
        """Check if a folder exists.

        Args:
            path: The path to the folder.

        Returns:
            bool: True if the folder exists, False otherwise.

        """

    @abstractmethod
    def file_exists(self, path: str) -> bool:
        """Check if a file exists.

        Args:
            path: The path to the file.

        Returns:
            bool: True if the file exists, False otherwise.

        """

    @abstractmethod
    def write(self, file_path: str, file_content: str, overwrite: bool = True) -> None:
        """Write content to a file.

        Args:
            file_path: The name of the file to write to.
            file_content: The content to write to the file.
            overwrite: If True, overwrite the file if it exists.

        """

    @abstractmethod
    def read(
        self, file_path: str, mode: str = "r", encoding: Optional[str] = None
    ) -> str:
        """Read content from a file.

        Args:
            file_path: The path to the file to read from.
            mode: The mode in which to open the file.
            encoding: The encoding to use for reading the file.

        Returns:
            str: The content of the file.

        """

    @abstractmethod
    def read_bytes(self, file_path: str) -> bytes:
        """Read binary content from a file.

        Args:
            file_path: The path to the file to read from.

        Returns:
            bytes: The binary content of the file.

        """

    @abstractmethod
    def ls(self, path: str, recursive: bool = False) -> list[str]:
        """List the contents of a directory.

        Args:
            path: The path to the directory.
            recursive: If True, list the contents recursively.

        Returns:
            list[str]: A list of the contents of the directory.

        """

    @abstractmethod
    def getcwd(self) -> str:
        """Get the current working directory.

        Returns:
            str: The current working directory.

        """

    @abstractmethod
    def telemetry_path_files(self, path: str) -> Path:
        """Get the path to the telemetry files.

        Args:
            path: The path to the telemetry directory.

        Returns:
            Path: The path object representing the telemetry files.

        """



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date, datetime
from zoneinfo import ZoneInfo

import numpy as np
import pandas as pd
import pandera as pa
import pytest

from snowflake.hypothesis_snowpark.constants import PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES
from snowflake.hypothesis_snowpark.datatype_mapper import (
    pandera_dtype_to_snowpark_dtype,
    pyspark_dtype_to_snowpark_dtype,
)
from snowflake.snowpark.types import (
    BooleanType,
    ByteType,
    DateType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    ShortType,
    StringType,
    TimestampTimeZone,
    TimestampType,
)
from snowflake.snowpark.types import (
    DataType as SnowparkDataType,
)


def test_pandera_dtype_to_snowpark_dtype_unsupported():
    with pytest.raises(
        TypeError,
        match="Unsupported Pandera data type: <class 'pandera.dtypes.Category'>",
    ):
        pandera_dtype_to_snowpark_dtype(pa.Category)


def test_pandera_dtype_to_snowpark_dtype_byte():
    column_type_mapping = [
        (pa.Column(pa.Int8), ByteType()),
        (pa.Column(pa.INT8), ByteType()),
        (pa.Column("int8"), ByteType()),
        (pa.Column(np.int8), ByteType()),
        (pa.Column(pd.Int8Dtype), ByteType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_short():
    column_type_mapping = [
        (pa.Column(pa.Int16), ShortType()),
        (pa.Column(pa.INT16), ShortType()),
        (pa.Column("int16"), ShortType()),
        (pa.Column(np.int16), ShortType()),
        (pa.Column(pd.Int16Dtype), ShortType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_integer():
    column_type_mapping = [
        (pa.Column(pa.Int32), IntegerType()),
        (pa.Column(pa.INT32), IntegerType()),
        (pa.Column("int32"), IntegerType()),
        (pa.Column(np.int32), IntegerType()),
        (pa.Column(pd.Int32Dtype), IntegerType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_long():
    column_type_mapping = [
        (pa.Column(pa.Int64), LongType()),
        (pa.Column(pa.INT64), LongType()),
        (pa.Column(pa.Int), LongType()),
        (pa.Column(int), LongType()),
        (pa.Column("int"), LongType()),
        (pa.Column("int64"), LongType()),
        (pa.Column(np.int64), LongType()),
        (pa.Column(np.int_), LongType()),
        (pa.Column(pd.Int64Dtype), LongType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_float():
    column_type_mapping = [
        (pa.Column(pa.Float32), FloatType()),
        (pa.Column("float32"), FloatType()),
        (pa.Column(np.float32), FloatType()),
        (pa.Column(pd.Float32Dtype), FloatType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_double():
    column_type_mapping = [
        (pa.Column(pa.Float64), DoubleType()),
        (pa.Column(float), DoubleType()),
        (pa.Column("float"), DoubleType()),
        (pa.Column("float64"), DoubleType()),
        (pa.Column(np.float64), DoubleType()),
        (pa.Column(pd.Float64Dtype), DoubleType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_string():
    column_type_mapping = [
        (pa.Column(pa.String), StringType()),
        (pa.Column(pa.STRING), StringType()),
        (pa.Column(str), StringType()),
        (pa.Column("str"), StringType()),
        (pa.Column("string"), StringType()),
        (pa.Column(np.str_), StringType()),
        (pa.Column(pd.StringDtype), StringType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_boolean():
    column_type_mapping = [
        (pa.Column(pa.Bool), BooleanType()),
        (pa.Column(pa.BOOL), BooleanType()),
        (pa.Column(bool), BooleanType()),
        (pa.Column("bool"), BooleanType()),
        (pa.Column("boolean"), BooleanType()),
        (pa.Column(np.bool_), BooleanType()),
        (pa.Column(pd.BooleanDtype), BooleanType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_timestamp():
    column_type_mapping = [
        (
            pa.Column(pd.DatetimeTZDtype(tz=ZoneInfo("Europe/Paris"))),
            TimestampType(TimestampTimeZone.TZ),
        ),
        (
            pa.Column("datetime64[ns, Europe/Paris]"),
            TimestampType(TimestampTimeZone.TZ),
        ),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_timestamp_ntz():
    column_type_mapping = [
        (pa.Column(pa.Timestamp), TimestampType(TimestampTimeZone.NTZ)),
        (pa.Column(pa.DateTime), TimestampType(TimestampTimeZone.NTZ)),
        (pa.Column(datetime), TimestampType(TimestampTimeZone.NTZ)),
        (pa.Column("datetime"), TimestampType(TimestampTimeZone.NTZ)),
        (pa.Column("datetime64[ns]"), TimestampType(TimestampTimeZone.NTZ)),
        (pa.Column(np.datetime64), TimestampType(TimestampTimeZone.NTZ)),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pandera_dtype_to_snowpark_dtype_timestamp_date():
    column_type_mapping = [
        (pa.Column(pa.Date), DateType()),
        (pa.Column(date), DateType()),
        (pa.Column("date"), DateType()),
    ]
    _validate_column_type_mapping(column_type_mapping)


def test_pyspark_dtype_to_snowpark_dtype_valid_types():
    for (
        pyspark_type,
        expected_snowpark_type,
    ) in PYSPARK_TO_SNOWPARK_SUPPORTED_TYPES.items():
        actual_snowpark_type = pyspark_dtype_to_snowpark_dtype(pyspark_type)
        assert isinstance(actual_snowpark_type, SnowparkDataType)
        assert actual_snowpark_type == expected_snowpark_type


def test_pyspark_dtype_to_snowpark_dtype_invalid_types():
    pyspark_type = "unknown_type"
    with pytest.raises(
        TypeError, match=f"Unsupported PySpark data type: {pyspark_type}"
    ):
        pyspark_dtype_to_snowpark_dtype(pyspark_type)


def _validate_column_type_mapping(
    column_type_mapping: list[tuple[pa.Column, SnowparkDataType]]
):
    for column, expected_snowpark_type in column_type_mapping:
        assert pandera_dtype_to_snowpark_dtype(column.dtype) == expected_snowpark_type



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import datetime
import hashlib
import inspect
import json
import os
import re

from contextlib import suppress
from enum import IntEnum
from functools import wraps
from pathlib import Path
from platform import python_version
from sys import platform
from typing import Any, Callable, Optional, TypeVar
from uuid import getnode

from snowflake.connector.description import PLATFORM as CONNECTOR_PLATFORM
from snowflake.connector.telemetry import TelemetryClient
from snowflake.hypothesis_snowpark.io_utils.io_file_manager import (
    get_io_file_manager,
)
from snowflake.snowpark import VERSION as SNOWPARK_VERSION
from snowflake.snowpark import dataframe as snowpark_dataframe
from snowflake.snowpark.session import Session


try:
    """
    The following imports are used to log telemetry events in the Snowflake Connector.
    """
    from snowflake.connector import (
        SNOWFLAKE_CONNECTOR_VERSION,
        time_util,
    )
    from snowflake.connector.constants import DIRS as SNOWFLAKE_DIRS
    from snowflake.connector.network import SnowflakeRestful
except Exception:
    """
    Set default import values for the Snowflake Connector when using snowpark-checkpoints in stored procedures.
    """
    SNOWFLAKE_CONNECTOR_VERSION = ""
    time_util = None
    SNOWFLAKE_DIRS = ""
    SnowflakeRestful = None


try:
    from pyspark.sql import dataframe as spark_dataframe

    def _is_spark_dataframe(df: Any) -> bool:
        return isinstance(df, spark_dataframe.DataFrame)

    def _get_spark_schema_types(df: spark_dataframe.DataFrame) -> list[str]:
        return [str(schema_type.dataType) for schema_type in df.schema.fields]

except Exception:

    def _is_spark_dataframe(df: Any):
        pass

    def _get_spark_schema_types(df: Any):
        pass


VERSION_VARIABLE_PATTERN = r"^__version__ = ['\"]([^'\"]*)['\"]"
VERSION_FILE_NAME = "__version__.py"


class TelemetryManager(TelemetryClient):
    def __init__(
        self, rest: Optional[SnowflakeRestful] = None, is_telemetry_enabled: bool = True
    ):
        """TelemetryManager class to log telemetry events."""
        super().__init__(rest)
        self.sc_folder_path = (
            Path(SNOWFLAKE_DIRS.user_config_path) / "snowpark-checkpoints-telemetry"
        )
        self.sc_sf_path_telemetry = "/telemetry/send"
        self.sc_flush_size = 25
        self.sc_is_enabled = is_telemetry_enabled
        self.sc_is_testing = self._sc_is_telemetry_testing()
        self.sc_memory_limit = 5 * 1024 * 1024
        self._sc_upload_local_telemetry()
        self.sc_log_batch = []
        self.sc_hypothesis_input_events = []
        self.sc_version = _get_version()
        if rest:
            atexit.register(self._sc_close_at_exit)

    def set_sc_output_path(self, path: Path) -> None:
        """Set the output path for testing.

        Args:
            path: path to write telemetry.

        """
        get_io_file_manager().mkdir(str(path), exist_ok=True)
        self.sc_folder_path = path

    def sc_log_error(
        self, event_name: str, parameters_info: Optional[dict] = None
    ) -> None:
        """Log an error telemetry event.

        Args:
            event_name (str): The name of the event.
            parameters_info (dict, optional): Additional parameters for the event. Defaults to None.

        """
        if event_name is not None:
            self._sc_log_telemetry(event_name, "error", parameters_info)

    def sc_log_info(
        self, event_name: str, parameters_info: Optional[dict] = None
    ) -> None:
        """Log an information telemetry event.

        Args:
            event_name (str): The name of the event.
            parameters_info (dict, optional): Additional parameters for the event. Defaults to None.

        """
        if event_name is not None:
            self._sc_log_telemetry(event_name, "info", parameters_info)

    def _sc_log_telemetry(
        self, event_name: str, event_type: str, parameters_info: Optional[dict] = None
    ) -> dict:
        """Log a telemetry event if is enabled.

        Args:
            event_name (str): The name of the event.
            event_type (str): The type of the event (e.g., "error", "info").
            parameters_info (dict, optional): Additional parameters for the event. Defaults to None.

        Returns:
            dict: The logged event.

        """
        if not self.sc_is_enabled:
            return {}
        event = _generate_event(
            event_name, event_type, parameters_info, self.sc_version
        )
        self._sc_add_log_to_batch(event)
        return event

    def _sc_add_log_to_batch(self, event: dict) -> None:
        """Add a log event to the batch. If the batch is full, send the events to the API.

        Args:
            event (dict): The event to add.

        """
        self.sc_log_batch.append(event)
        if self.sc_is_testing:
            self._sc_write_telemetry(self.sc_log_batch)
            self.sc_log_batch = []
            return

        if len(self.sc_log_batch) >= self.sc_flush_size:
            self.sc_send_batch(self.sc_log_batch)
            self.sc_log_batch = []

    def sc_send_batch(self, to_sent: list) -> bool:
        """Send a request to the API to upload the events. If not have connection, write the events to local folder.

        Args:
            to_sent (list): The batch of events to send.

        Returns:
            bool: True if the batch was sent successfully, False otherwise.

        """
        if not self.sc_is_enabled:
            return False
        if self._rest is None:
            self._sc_write_telemetry(to_sent)
            self.sc_log_batch = []
            return False
        if to_sent == []:
            return False
        body = {"logs": to_sent}
        ret = self._rest.request(
            self.sc_sf_path_telemetry,
            body=body,
            method="post",
            client=None,
            timeout=5,
        )
        if not ret.get("success"):
            self._sc_write_telemetry(to_sent)
            self.sc_log_batch = []
            return False
        return True

    def _sc_write_telemetry(self, batch: list) -> None:
        """Write telemetry events to local folder. If the folder is full, free up space for the new events.

        Args:
            batch (list): The batch of events to write.

        """
        try:
            get_io_file_manager().mkdir(str(self.sc_folder_path), exist_ok=True)
            for event in batch:
                message = event.get("message")
                if message is not None:
                    file_path = (
                        self.sc_folder_path
                        / f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S-%f")}'
                        f'_telemetry_{message.get("type")}.json'
                    )
                    json_content = self._sc_validate_folder_space(event)
                    get_io_file_manager().write(str(file_path), json_content)
        except Exception:
            pass

    def _sc_validate_folder_space(self, event: dict) -> str:
        """Validate and manage folder space for the new telemetry events.

        Args:
            event (dict): The event to validate.

        Returns:
            str: The JSON content of the event.

        """
        json_content = json.dumps(event, indent=4, sort_keys=True)
        new_json_file_size = len(json_content.encode("utf-8"))
        telemetry_folder = self.sc_folder_path
        folder_size = _get_folder_size(telemetry_folder)
        if folder_size + new_json_file_size > self.sc_memory_limit:
            _free_up_space(telemetry_folder, self.sc_memory_limit - new_json_file_size)
        return json_content

    def _sc_upload_local_telemetry(self) -> None:
        """Send a request to the API to upload the local telemetry events."""
        if not self.sc_is_enabled or self.sc_is_testing or not self._rest:
            return
        batch = []
        for file in get_io_file_manager().ls(f"{self.sc_folder_path}/*.json"):
            json_content = get_io_file_manager().read(file)
            data_dict = json.loads(json_content)
            batch.append(data_dict)
        if batch == []:
            return
        body = {"logs": batch}
        ret = self._rest.request(
            self.sc_sf_path_telemetry,
            body=body,
            method="post",
            client=None,
            timeout=5,
        )
        if ret.get("success"):
            for file_path in get_io_file_manager().ls(f"{self.sc_folder_path}/*.json"):
                file = get_io_file_manager().telemetry_path_files(file_path)
                file.unlink()

    def _sc_is_telemetry_testing(self) -> bool:
        is_testing = os.getenv("SNOWPARK_CHECKPOINTS_TELEMETRY_TESTING") == "true"
        if is_testing:
            local_telemetry_path = (
                Path(get_io_file_manager().getcwd())
                / "snowpark-checkpoints-output"
                / "telemetry"
            )
            self.set_sc_output_path(local_telemetry_path)
            self.sc_is_enabled = True
        return is_testing

    def _sc_is_telemetry_manager(self) -> bool:
        """Check if the class is telemetry manager.

        Returns:
            bool: True if the class is telemetry manager, False otherwise.

        """
        return True

    def sc_is_hypothesis_event_logged(self, event_name: tuple[str, int]) -> bool:
        """Check if a Hypothesis event is logged.

        Args:
            event_name (tuple[str, int]): A tuple containing the name of the event and an integer identifier
            (0 for info, 1 for error).

        Returns:
            bool: True if the event is logged, False otherwise.

        """
        return event_name in self.sc_hypothesis_input_events

    def _sc_close(self) -> None:
        """Close the telemetry manager and upload collected events.

        This function closes the telemetry manager, uploads any collected events,
        and performs any necessary cleanup to ensure no data is lost.
        """
        atexit.unregister(self._sc_close_at_exit)
        if self.sc_log_batch and self.sc_is_enabled and not self.sc_is_testing:
            self.sc_send_batch(self.sc_log_batch)

    def _sc_close_at_exit(self) -> None:
        """Close the telemetry manager at exit and upload collected events.

        This function ensures that the telemetry manager is closed and all collected events
        are uploaded when the program exits, preventing data loss.
        """
        with suppress(Exception):
            self._sc_close()


def _generate_event(
    event_name: str,
    event_type: str,
    parameters_info: Optional[dict] = None,
    sc_version: Optional[str] = None,
) -> dict:
    """Generate a telemetry event.

    Args:
        event_name (str): The name of the event.
        event_type (str): The type of the event (e.g., "error", "info").
        parameters_info (dict, optional): Additional parameters for the event. Defaults to None.
        sc_version (str, optional): The version of the package. Defaults to None.

    Returns:
        dict: The generated event.

    """
    metadata = _get_metadata()
    if sc_version is not None:
        metadata["snowpark_checkpoints_version"] = sc_version
    message = {
        "event_type": event_type,
        "type": "snowpark-checkpoints",
        "event_name": event_name,
        "driver_type": "PythonConnector",
        "driver_version": SNOWFLAKE_CONNECTOR_VERSION,
        "metadata": metadata,
        "data": json.dumps(parameters_info or {}),
    }
    timestamp = time_util.get_time_millis()
    event_base = {"message": message, "timestamp": str(timestamp)}

    return event_base


def _get_metadata() -> dict:
    """Get metadata for telemetry events.

    Returns:
        dict: The metadata including OS version, Python version, and device ID.

    """
    return {
        "os_version": platform,
        "python_version": python_version(),
        "snowpark_version": ".".join(str(x) for x in SNOWPARK_VERSION if x is not None),
        "device_id": _get_unique_id(),
    }


def _get_version() -> Optional[str]:
    """Get the version of the package.

    Returns:
        str: The version of the package.

    """
    try:
        directory_levels_up = 1
        project_root = Path(__file__).resolve().parents[directory_levels_up]
        version_file_path = project_root / VERSION_FILE_NAME
        content = get_io_file_manager().read(str(version_file_path))
        version_match = re.search(VERSION_VARIABLE_PATTERN, content, re.MULTILINE)
        if version_match:
            return version_match.group(1)
        return None
    except Exception:
        return None


def _get_folder_size(folder_path: Path) -> int:
    """Get the size of a folder. Only considers JSON files.

    Args:
        folder_path (Path): The path to the folder.

    Returns:
        int: The size of the folder in bytes.

    """
    sum_size = 0
    for f in get_io_file_manager().ls(f"{folder_path}/*.json"):
        sum_size += get_io_file_manager().telemetry_path_files(f).stat().st_size
    return sum_size


def _free_up_space(folder_path: Path, max_size: int) -> None:
    """Free up space in a folder by deleting the oldest files. Only considers JSON files.

    Args:
        folder_path (Path): The path to the folder.
        max_size (int): The maximum allowed size of the folder in bytes.

    """
    files = sorted(
        get_io_file_manager().ls(f"{folder_path}/*.json"),
        key=lambda f: f.stat().st_mtime,
    )
    current_size = _get_folder_size(folder_path)
    for file_path in files:
        file = get_io_file_manager().telemetry_path_files(file_path)
        if current_size <= max_size:
            break
        current_size -= file.stat().st_size
        file.unlink()


def _get_unique_id() -> str:
    """Get a unique device ID. The ID is generated based on the hashed MAC address.

    Returns:
        str: The hashed device ID.

    """
    node_id_str = str(getnode())
    hashed_id = hashlib.sha256(node_id_str.encode()).hexdigest()
    return hashed_id


def get_telemetry_manager() -> TelemetryManager:
    """Get the telemetry manager.

    Returns:
        TelemetryManager: The telemetry manager.

    """
    try:
        connection = Session.builder.getOrCreate().connection
        if not hasattr(connection._telemetry, "_sc_is_telemetry_manager"):
            connection._telemetry = TelemetryManager(
                connection._rest, connection.telemetry_enabled
            )
        return connection._telemetry
    except Exception:
        telemetry_manager = TelemetryManager(None, is_telemetry_enabled=True)
        telemetry_manager.sc_flush_size = 1
        return telemetry_manager


def get_snowflake_schema_types(df: snowpark_dataframe.DataFrame) -> list[str]:
    """Extract the data types of the schema fields from a Snowflake DataFrame.

    Args:
        df (snowpark_dataframe.DataFrame): The Snowflake DataFrame.

    Returns:
        list[str]: A list of data type names of the schema fields.

    """
    return [str(schema_type.datatype) for schema_type in df.schema.fields]


def _is_snowpark_dataframe(df: Any) -> bool:
    """Check if the given dataframe is a Snowpark dataframe.

    Args:
        df: The dataframe to check.

    Returns:
        bool: True if the dataframe is a Snowpark dataframe, False otherwise.

    """
    return isinstance(df, snowpark_dataframe.DataFrame)


def get_load_json(json_schema: str) -> dict:
    """Load and parse a JSON schema file.

    Args:
        json_schema (str): The path to the JSON schema file.

    Returns:
        dict: The parsed JSON content.

    Raises:
        ValueError: If there is an error reading or parsing the JSON file.

    """
    try:
        file_content = get_io_file_manager().read(json_schema, encoding="utf-8")
        return json.loads(file_content)
    except (OSError, json.JSONDecodeError) as e:
        raise ValueError(f"Error reading JSON schema file: {e}") from None


def _is_in_stored_procedure() -> bool:
    """Check if the code is running in a stored procedure.

    Returns:
        bool: True if the code is running in a stored procedure, False otherwise.

    """
    return CONNECTOR_PLATFORM == "XP"


def extract_parameters(
    func: Callable, args: tuple, kwargs: dict, params_list: Optional[list[str]]
) -> dict:
    """Extract parameters from the function arguments.

    Args:
        func (Callable): The function being decorated.
        args (tuple): The positional arguments passed to the function.
        kwargs (dict): The keyword arguments passed to the function.
        params_list (list[str]): The list of parameters to extract.

    Returns:
        dict: A dictionary of extracted parameters.

    """
    parameters = inspect.signature(func).parameters
    param_data = {}
    if params_list:
        for _, param in enumerate(params_list):
            if len(args) > 0:
                index = list(parameters.keys()).index(param)
                param_data[param] = args[index]
            else:
                if kwargs[param]:
                    param_data[param] = kwargs[param]
    return param_data


def check_dataframe_schema_event(
    telemetry_data: dict, param_data: dict
) -> tuple[str, dict]:
    """Handle telemetry event for checking dataframe schema.

    Args:
        telemetry_data (dict): The telemetry data dictionary.
        param_data (dict): The parameter data dictionary.

    Returns:
        tuple: A tuple containing the event name and telemetry data.

    """
    telemetry_data[MODE_KEY] = CheckpointMode.SCHEMA.value
    try:
        telemetry_data[STATUS_KEY] = param_data.get(STATUS_KEY)
        pandera_schema = param_data.get(PANDERA_SCHEMA_PARAM_NAME)
        schema_types = []
        for schema_type in pandera_schema.columns.values():
            if schema_type.dtype is not None:
                schema_types.append(str(schema_type.dtype))
        if schema_types:
            telemetry_data[SCHEMA_TYPES_KEY] = schema_types
        return DATAFRAME_VALIDATOR_SCHEMA, telemetry_data
    except Exception:
        if param_data.get(STATUS_KEY):
            telemetry_data[STATUS_KEY] = param_data.get(STATUS_KEY)
        pandera_schema = param_data.get(PANDERA_SCHEMA_PARAM_NAME)
        if pandera_schema:
            schema_types = []
            for schema_type in pandera_schema.columns.values():
                if schema_type.dtype is not None:
                    schema_types.append(str(schema_type.dtype))
            if schema_types:
                telemetry_data[SCHEMA_TYPES_KEY] = schema_types
        return DATAFRAME_VALIDATOR_ERROR, telemetry_data


def check_output_or_input_schema_event(
    telemetry_data: dict, param_data: dict
) -> tuple[str, dict]:
    """Handle telemetry event for checking output or input schema.

    Args:
        telemetry_data (dict): The telemetry data dictionary.
        param_data (dict): The parameter data dictionary.

    Returns:
        tuple: A tuple containing the event name and telemetry data.

    """
    try:
        pandera_schema = param_data.get(PANDERA_SCHEMA_PARAM_NAME)
        schema_types = []
        for schema_type in pandera_schema.columns.values():
            if schema_type.dtype is not None:
                schema_types.append(str(schema_type.dtype))
        if schema_types:
            telemetry_data[SCHEMA_TYPES_KEY] = schema_types
        return DATAFRAME_VALIDATOR_SCHEMA, telemetry_data
    except Exception:
        return DATAFRAME_VALIDATOR_ERROR, telemetry_data


def collect_dataframe_checkpoint_mode_schema_event(
    telemetry_data: dict, param_data: dict
) -> tuple[str, dict]:
    """Handle telemetry event for collecting dataframe checkpoint mode schema.

    Args:
        telemetry_data (dict): The telemetry data dictionary.
        param_data (dict): The parameter data dictionary.

    Returns:
        tuple: A tuple containing the event name and telemetry data.

    """
    telemetry_data[MODE_KEY] = CheckpointMode.SCHEMA.value
    try:
        schema_types = param_data.get("column_type_dict")
        telemetry_data[SCHEMA_TYPES_KEY] = [
            schema_types[schema_type].dataType.typeName()
            for schema_type in schema_types
        ]
        return DATAFRAME_COLLECTION_SCHEMA, telemetry_data
    except Exception:
        return DATAFRAME_COLLECTION_ERROR, telemetry_data


def collect_dataframe_checkpoint_mode_dataframe_event(
    telemetry_data: dict, param_data: dict
) -> tuple[str, dict]:
    """Handle telemetry event for collecting dataframe checkpoint mode dataframe.

    This function processes telemetry data for a dataframe checkpoint mode event. It updates the telemetry data
    with the mode and schema types of the Spark DataFrame being collected.

    Args:
        telemetry_data (dict): The telemetry data dictionary to be updated.
        param_data (dict): The parameter data dictionary containing the DataFrame information.

    Returns:
        tuple: A tuple containing the event name and the updated telemetry data dictionary.

    """
    telemetry_data[MODE_KEY] = CheckpointMode.DATAFRAME.value
    try:
        if _is_spark_dataframe(param_data.get(DF_PARAM_NAME)):
            telemetry_data[SPARK_SCHEMA_TYPES_KEY] = _get_spark_schema_types(
                param_data.get(DF_PARAM_NAME)
            )
            return DATAFRAME_COLLECTION_DF, telemetry_data
    except Exception:
        return DATAFRAME_COLLECTION_ERROR, telemetry_data


def assert_return_event(telemetry_data: dict, param_data: dict) -> tuple[str, dict]:
    """Handle telemetry event for asserting return values.

    Args:
        telemetry_data (dict): The telemetry data dictionary.
        param_data (dict): The parameter data dictionary.

    Returns:
        tuple: A tuple containing the event name and telemetry data.

    """
    if param_data.get(STATUS_KEY) is not None:
        telemetry_data[STATUS_KEY] = param_data.get(STATUS_KEY, None)
    try:
        if _is_snowpark_dataframe(
            param_data.get(SNOWPARK_RESULTS_PARAM_NAME)
        ) and _is_spark_dataframe(param_data.get(SPARK_RESULTS_PARAM_NAME)):
            telemetry_data[SNOWFLAKE_SCHEMA_TYPES_KEY] = get_snowflake_schema_types(
                param_data.get(SNOWPARK_RESULTS_PARAM_NAME)
            )
            telemetry_data[SPARK_SCHEMA_TYPES_KEY] = _get_spark_schema_types(
                param_data.get(SPARK_RESULTS_PARAM_NAME)
            )
            return DATAFRAME_VALIDATOR_MIRROR, telemetry_data
        else:
            return VALUE_VALIDATOR_MIRROR, telemetry_data
    except Exception:
        if _is_snowpark_dataframe(param_data.get(SNOWPARK_RESULTS_PARAM_NAME)):
            telemetry_data[SNOWFLAKE_SCHEMA_TYPES_KEY] = get_snowflake_schema_types(
                param_data.get(SNOWPARK_RESULTS_PARAM_NAME)
            )
        if _is_spark_dataframe(param_data.get(SPARK_RESULTS_PARAM_NAME)):
            telemetry_data[SPARK_SCHEMA_TYPES_KEY] = _get_spark_schema_types(
                param_data.get(SPARK_RESULTS_PARAM_NAME)
            )
        return DATAFRAME_VALIDATOR_ERROR, telemetry_data


def dataframe_strategy_event(
    telemetry_data: dict, param_data: dict, telemetry_m: TelemetryManager
) -> tuple[Optional[str], Optional[dict]]:
    """Handle telemetry event for dataframe strategy.

    Args:
        telemetry_data (dict): The telemetry data dictionary.
        param_data (dict): The parameter data dictionary.
        telemetry_m (TelemetryManager): The telemetry manager.

    Returns:
        tuple: A tuple containing the event name and telemetry data.

    """
    try:
        test_function_name = inspect.stack()[2].function
        is_logged = telemetry_m.sc_is_hypothesis_event_logged((test_function_name, 0))
        if not is_logged:
            schema_param = param_data.get(DATAFRAME_STRATEGY_SCHEMA_PARAM_NAME)
            if isinstance(schema_param, str):
                json_data = get_load_json(schema_param)["custom_data"]["columns"]
                telemetry_data[SCHEMA_TYPES_KEY] = [
                    column["type"] for column in json_data
                ]
            else:
                schema_types = []
                for schema_type in schema_param.columns.values():
                    if schema_type.dtype is not None:
                        schema_types.append(str(schema_type.dtype))
                if schema_types:
                    telemetry_data[SCHEMA_TYPES_KEY] = schema_types
            telemetry_m.sc_hypothesis_input_events.append((test_function_name, 0))
            if None in telemetry_data[SCHEMA_TYPES_KEY]:
                telemetry_m.sc_log_error(HYPOTHESIS_INPUT_SCHEMA_ERROR, telemetry_data)
            else:
                telemetry_m.sc_log_info(HYPOTHESIS_INPUT_SCHEMA, telemetry_data)
            telemetry_m.sc_send_batch(telemetry_m.sc_log_batch)
        return None, None
    except Exception:
        test_function_name = inspect.stack()[2].function
        is_logged = telemetry_m.sc_is_hypothesis_event_logged((test_function_name, 1))
        if not is_logged:
            telemetry_m.sc_hypothesis_input_events.append((test_function_name, 0))
            telemetry_m.sc_log_error(HYPOTHESIS_INPUT_SCHEMA_ERROR, telemetry_data)
            telemetry_m.sc_send_batch(telemetry_m.sc_log_batch)
        return None, None


def compare_data_event(telemetry_data: dict, param_data: dict) -> tuple[str, dict]:
    """Handle telemetry event for comparing data.

    This function processes telemetry data for a data comparison event. It updates the telemetry data
    with the mode, status, and schema types of the Snowflake DataFrame being compared.

    Args:
        telemetry_data (dict): The telemetry data dictionary to be updated.
        param_data (dict): The parameter data dictionary containing the DataFrame and status information.

    Returns:
        tuple: A tuple containing the event name and the updated telemetry data dictionary.

    """
    telemetry_data[MODE_KEY] = CheckpointMode.DATAFRAME.value
    telemetry_data[STATUS_KEY] = param_data.get(STATUS_KEY, None)
    try:
        telemetry_data[SCHEMA_TYPES_KEY] = get_snowflake_schema_types(
            param_data.get("df")
        )
        return DATAFRAME_VALIDATOR_DF, telemetry_data
    except Exception:
        return DATAFRAME_VALIDATOR_ERROR, telemetry_data


def handle_result(
    func_name: str,
    result: Any,
    param_data: dict,
    multiple_return: bool,
    telemetry_m: TelemetryManager,
    return_indexes: Optional[list[tuple[str, int]]] = None,
) -> tuple[Optional[str], Optional[dict]]:
    """Handle the result of the function and collect telemetry data.

    Args:
        func_name (str): The name of the function.
        result: The result of the function.
        param_data (dict): The extracted parameters.
        multiple_return (bool): Whether the function returns multiple values.
        telemetry_m (TelemetryManager): The telemetry manager.
        return_indexes (list[tuple[str, int]]): The list of return values to report. Defaults to None.

    Returns:
        tuple: A tuple containing the event name (str) and telemetry data (dict).

    """
    if result is not None and return_indexes is not None:
        if multiple_return:
            for name, index in return_indexes:
                param_data[name] = result[index]
        else:
            param_data[return_indexes[0][0]] = result[return_indexes[0][1]]

    telemetry_data = {
        FUNCTION_KEY: func_name,
    }

    telemetry_event = None
    data = None
    if func_name == "_check_dataframe_schema":
        telemetry_event, data = check_dataframe_schema_event(telemetry_data, param_data)
    elif func_name in ["check_output_schema", "check_input_schema"]:
        telemetry_event, data = check_output_or_input_schema_event(
            telemetry_data, param_data
        )
    if func_name == "_compare_data":
        telemetry_event, data = compare_data_event(telemetry_data, param_data)
    elif func_name == "_collect_dataframe_checkpoint_mode_schema":
        telemetry_event, data = collect_dataframe_checkpoint_mode_schema_event(
            telemetry_data, param_data
        )
    elif func_name == "_collect_dataframe_checkpoint_mode_dataframe":
        telemetry_event, data = collect_dataframe_checkpoint_mode_dataframe_event(
            telemetry_data, param_data
        )
    elif func_name == "_assert_return":
        telemetry_event, data = assert_return_event(telemetry_data, param_data)
    elif func_name == "dataframe_strategy":
        telemetry_event, data = dataframe_strategy_event(
            telemetry_data, param_data, telemetry_m
        )
    return telemetry_event, data


fn = TypeVar("fn", bound=Callable)


def report_telemetry(
    params_list: list[str] = None,
    return_indexes: list[tuple[str, int]] = None,
    multiple_return: bool = False,
) -> Callable[[fn], fn]:
    """Report telemetry events for a function.

    Args:
        params_list (list[str], optional): The list of parameters to report. Defaults to None.
        return_indexes (list[tuple[str, int]], optional): The list of return values to report. Defaults to None.
        multiple_return (bool, optional): Whether the function returns multiple values. Defaults to False.

    Returns:
        Callable[[fn], fn]: The decorator function.

    """

    def report_telemetry_decorator(func):
        func_name = func.__name__

        @wraps(func)
        def wrapper(*args, **kwargs):
            func_exception = None
            result = None
            try:
                result = func(*args, **kwargs)
            except Exception as err:
                func_exception = err

            if (
                os.getenv("SNOWPARK_CHECKPOINTS_TELEMETRY_ENABLED") == "false"
                or _is_in_stored_procedure()
            ):
                return result
            telemetry_event = None
            data = None
            telemetry_m = None
            try:
                param_data = extract_parameters(func, args, kwargs, params_list)
                telemetry_m = get_telemetry_manager()
                telemetry_event, data = handle_result(
                    func_name,
                    result,
                    param_data,
                    multiple_return,
                    telemetry_m,
                    return_indexes,
                )
            except Exception:
                pass
            finally:
                if func_exception is not None:
                    if telemetry_m is not None:
                        telemetry_m.sc_log_error(telemetry_event, data)
                    raise func_exception
                if telemetry_m is not None:
                    telemetry_m.sc_log_info(telemetry_event, data)

            return result

        return wrapper

    return report_telemetry_decorator


# Constants for telemetry
DATAFRAME_COLLECTION_SCHEMA = "DataFrame_Collection_Schema"
DATAFRAME_COLLECTION_DF = "DataFrame_Collection_DF"
DATAFRAME_VALIDATOR_MIRROR = "DataFrame_Validator_Mirror"
VALUE_VALIDATOR_MIRROR = "Value_Validator_Mirror"
DATAFRAME_VALIDATOR_SCHEMA = "DataFrame_Validator_Schema"
DATAFRAME_VALIDATOR_DF = "DataFrame_Validator_DF"
HYPOTHESIS_INPUT_SCHEMA = "Hypothesis_Input_Schema"
DATAFRAME_COLLECTION_ERROR = "DataFrame_Collection_Error"
DATAFRAME_VALIDATOR_ERROR = "DataFrame_Validator_Error"
HYPOTHESIS_INPUT_SCHEMA_ERROR = "Hypothesis_Input_Schema_Error"

FUNCTION_KEY = "function"
STATUS_KEY = "status"
SCHEMA_TYPES_KEY = "schema_types"
ERROR_KEY = "error"
MODE_KEY = "mode"
SNOWFLAKE_SCHEMA_TYPES_KEY = "snowflake_schema_types"
SPARK_SCHEMA_TYPES_KEY = "spark_schema_types"

DATAFRAME_STRATEGY_SCHEMA_PARAM_NAME = "schema"
PANDERA_SCHEMA_PARAM_NAME = "pandera_schema"
SNOWPARK_RESULTS_PARAM_NAME = "snowpark_results"
SPARK_RESULTS_PARAM_NAME = "spark_results"
DF_PARAM_NAME = "df"


class CheckpointMode(IntEnum):
    SCHEMA = 1
    DATAFRAME = 2



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import logging

from collections.abc import Generator
from importlib import import_module
from typing import Any, Union

import pytest


TOP_LEVEL_LOGGER_NAME = "snowflake.hypothesis_snowpark"

LoggerDict = dict[str, Union[logging.Logger, logging.PlaceHolder]]
LoggerGenerator = Generator[logging.Logger, Any, None]


@pytest.fixture(name="registered_loggers", scope="module")
def fixture_registered_loggers() -> LoggerDict:
    """Return a dictionary with all the registered loggers."""
    # We need to import the snowflake.hypothesis_snowpark module
    # to ensure the __init__.py file is executed.
    import_module("snowflake.hypothesis_snowpark")

    return logging.getLogger().manager.loggerDict


@pytest.fixture(name="top_level_logger", scope="module")
def fixture_top_level_logger(registered_loggers: LoggerDict) -> LoggerGenerator:
    """Return the top-level logger of the snowpark-checkpoints-hypothesis package."""
    logger = registered_loggers.get(TOP_LEVEL_LOGGER_NAME, None)
    assert isinstance(logger, logging.Logger)

    # Save the original state of the logger
    original_handlers = logger.handlers
    original_propagate = logger.propagate
    original_level = logger.level
    original_filters = logger.filters

    yield logger

    # Restore the original state of the logger
    logger.handlers = original_handlers
    logger.propagate = original_propagate
    logger.setLevel(original_level)
    logger.filters = original_filters


def test_loggers_exists(registered_loggers: LoggerDict):
    """Validates that all the snowflake.hypothesis_snowpark loggers exist."""
    logger_names = {
        "snowflake.hypothesis_snowpark",
        "snowflake.hypothesis_snowpark.custom_strategies",
        "snowflake.hypothesis_snowpark.strategies",
        "snowflake.hypothesis_snowpark.strategies_utils",
        "snowflake.hypothesis_snowpark.strategy_register",
    }

    for logger_name in logger_names:
        logger = registered_loggers.get(logger_name, None)
        assert logger is not None
        assert isinstance(logger, logging.Logger)


def test_top_level_logger_has_null_handler(top_level_logger: LoggerGenerator):
    """Validates that the top-level logger has a single handler and that it is a NullHandler."""
    assert isinstance(top_level_logger, logging.Logger)
    assert len(top_level_logger.handlers) == 1
    assert isinstance(top_level_logger.handlers[0], logging.NullHandler)


def test_top_level_logger_default_log_level(top_level_logger: LoggerGenerator):
    """Validates that the default log level of the top-level logger is logging.WARNING."""
    assert isinstance(top_level_logger, logging.Logger)
    assert top_level_logger.getEffectiveLevel() == logging.WARNING


def test_child_logger_inheritance(
    registered_loggers: LoggerDict, top_level_logger: LoggerGenerator
):
    """Validates the inheritance of the loggers."""
    child_logger = registered_loggers.get(f"{TOP_LEVEL_LOGGER_NAME}.strategies", None)
    assert isinstance(child_logger, logging.Logger)
    assert child_logger.parent == top_level_logger


def test_log_propagation(
    registered_loggers: LoggerDict, top_level_logger: LoggerGenerator
):
    """Validates that log messages are propagated from a child logger to the top-level logger."""
    assert isinstance(top_level_logger, logging.Logger)

    stream_handler = logging.StreamHandler(io.StringIO())
    top_level_logger.addHandler(stream_handler)

    child_logger = registered_loggers.get(f"{TOP_LEVEL_LOGGER_NAME}.strategies", None)
    assert isinstance(child_logger, logging.Logger)
    assert len(child_logger.handlers) == 0

    child_logger.warning("Test message")
    assert "Test message" in stream_handler.stream.getvalue()


def test_null_handler_supresses_output(
    capsys: pytest.CaptureFixture[str], top_level_logger: LoggerGenerator
):
    """Validates that NullHandler suppresses output to stderr."""
    assert isinstance(top_level_logger, logging.Logger)
    top_level_logger.propagate = False
    top_level_logger.error("This should not appear in stderr")
    captured = capsys.readouterr()
    assert captured.err == ""



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import random

from pathlib import Path

import numpy as np
import pandas as pd
import pandera as pa
import pytest

from snowflake.hypothesis_snowpark.strategies_utils import (
    apply_custom_null_values,
    generate_snowpark_dataframe,
    generate_snowpark_schema,
    load_json_schema,
    replace_surrogate_chars,
    temporary_random_seed,
)
from snowflake.snowpark import DataFrame, Row, Session
from snowflake.snowpark.types import (
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
)


LOGGER_NAME = "snowflake.hypothesis_snowpark.strategies_utils"


def test_load_json_schema_valid_file(tmp_path: Path):
    json_content = {"key": "value"}
    json_file = tmp_path / "valid_file.json"
    json_file.write_text(json.dumps(json_content))
    result = load_json_schema(str(json_file))
    assert result == json_content


def test_load_json_schema_invalid_path():
    json_file = "invalid_path.json"
    with pytest.raises(ValueError, match=f"Invalid JSON schema path: {json_file}"):
        load_json_schema(json_file)


def test_load_json_schema_invalid_json(tmp_path: Path):
    invalid_json_file = tmp_path / "invalid_json.json"
    invalid_json_file.write_text("{invalid_json}")
    with pytest.raises(ValueError, match="Error reading JSON schema file:"):
        load_json_schema(str(invalid_json_file))


def test_apply_custom_null_values_no_nulls():
    data = {"col1": [1, 2, 3, 4, 5], "col2": ["a", "b", "c", "d", "e"]}
    pandas_df = pd.DataFrame(data)
    custom_data = {
        "columns": [
            {"name": "col1", "rows_null_count": 0, "rows_count": 5},
            {"name": "col2", "rows_null_count": 0, "rows_count": 5},
        ]
    }
    result_df = apply_custom_null_values(pandas_df, custom_data)
    assert result_df.isnull().sum().sum() == 0


def test_apply_custom_null_values_some_nulls():
    data = {"col1": [1, 2, 3, 4, 5], "col2": ["a", "b", "c", "d", "e"]}
    pandas_df = pd.DataFrame(data)
    custom_data = {
        "columns": [
            {"name": "col1", "rows_null_count": 2, "rows_count": 5},
            {"name": "col2", "rows_null_count": 1, "rows_count": 5},
        ]
    }
    result_df = apply_custom_null_values(pandas_df, custom_data)
    assert result_df["col1"].isnull().sum() == 2
    assert result_df["col2"].isnull().sum() == 1


def test_apply_custom_null_values_all_nulls():
    data = {"col1": [1, 2, 3, 4, 5], "col2": ["a", "b", "c", "d", "e"]}
    pandas_df = pd.DataFrame(data)
    custom_data = {
        "columns": [
            {"name": "col1", "rows_null_count": 5, "rows_count": 5},
            {"name": "col2", "rows_null_count": 5, "rows_count": 5},
        ]
    }
    result_df = apply_custom_null_values(pandas_df, custom_data)
    assert result_df["col1"].isnull().sum() == 5
    assert result_df["col2"].isnull().sum() == 5


def test_apply_custom_null_values_invalid_column(caplog: pytest.LogCaptureFixture):
    data = {"col1": [1, 2, 3, 4, 5], "col2": ["a", "b", "c", "d", "e"]}
    pandas_df = pd.DataFrame(data)
    custom_data = {"columns": [{"name": "col3", "rows_null_count": 2, "rows_count": 5}]}
    with caplog.at_level(level=logging.WARNING, logger=LOGGER_NAME):
        result_df = apply_custom_null_values(pandas_df, custom_data)
    assert result_df["col1"].isnull().sum() == 0
    assert result_df["col2"].isnull().sum() == 0
    assert "col3" not in result_df.columns
    assert "'col3' not found in DataFrame" in caplog.text


def test_generate_snowpark_schema_no_custom_data():
    pandera_schema = pa.DataFrameSchema(
        {
            "name": pa.Column(pa.String, nullable=False),
            "age": pa.Column(pa.Int, nullable=True),
        }
    )

    expected_schema = StructType(
        [
            StructField("name", StringType(), False),
            StructField("age", LongType(), True),
        ]
    )

    snowpark_schema = generate_snowpark_schema(pandera_schema)
    assert snowpark_schema == expected_schema


def test_generate_snowpark_schema_with_custom_data():
    pandera_schema = pa.DataFrameSchema(
        {
            "name": pa.Column(pa.String, nullable=False),
            "age": pa.Column(pa.Int, nullable=True),
        }
    )

    custom_data = {
        "columns": [
            {"name": "name", "type": "string"},
            {"name": "age", "type": "integer"},
        ]
    }

    expected_schema = StructType(
        [
            StructField("name", StringType(), False),
            StructField("age", IntegerType(), True),
        ]
    )

    snowpark_schema = generate_snowpark_schema(pandera_schema, custom_data)
    assert snowpark_schema == expected_schema


def test_generate_snowpark_schema_missing_column_in_custom_data():
    pandera_schema = pa.DataFrameSchema(
        {
            "name": pa.Column(pa.String, nullable=False),
            "age": pa.Column(pa.Int, nullable=True),
        }
    )

    custom_data = {
        "columns": [
            {"name": "name", "type": "string"},
        ]
    }

    with pytest.raises(ValueError, match="Column 'age' is missing from custom_data"):
        generate_snowpark_schema(pandera_schema, custom_data)


def test_generate_snowpark_schema_missing_type_in_custom_data():
    pandera_schema = pa.DataFrameSchema(
        {
            "name": pa.Column(pa.String, nullable=False),
            "age": pa.Column(pa.Int, nullable=True),
        }
    )

    custom_data = {
        "columns": [
            {"name": "name", "type": "string"},
            {"name": "age"},
        ]
    }

    with pytest.raises(
        ValueError, match="Type for column 'age' is missing from custom_data"
    ):
        generate_snowpark_schema(pandera_schema, custom_data)


def test_generate_snowpark_schema_invalid_pyspark_type():
    pandera_schema = pa.DataFrameSchema(
        {
            "name": pa.Column(pa.String, nullable=False),
            "age": pa.Column(pa.Int, nullable=True),
        }
    )

    custom_data = {
        "columns": [
            {"name": "name", "type": "string"},
            {"name": "age", "type": "unknown_type"},
        ]
    }

    with pytest.raises(TypeError, match="Unsupported PySpark data type: unknown_type"):
        generate_snowpark_schema(pandera_schema, custom_data)


def test_generate_snowpark_dataframe_with_valid_data(
    session: Session, caplog: pytest.LogCaptureFixture
):
    pandas_df = pd.DataFrame({"col1": [1, 2], "col2": ["a", "b"]})
    snowpark_schema = StructType(
        [
            StructField("col1", IntegerType(), nullable=False),
            StructField("col2", StringType(), nullable=False),
        ]
    )
    with caplog.at_level(level=logging.INFO, logger=LOGGER_NAME):
        result = generate_snowpark_dataframe(pandas_df, snowpark_schema, session)
    assert isinstance(result, DataFrame)
    assert result.collect() == [Row(COL1=1, COL2="a"), Row(COL1=2, COL2="b")]
    for substring in [str(snowpark_schema), str(len(pandas_df))]:
        assert substring in caplog.text


def test_generate_snowpark_dataframe_with_empty_dataframe(
    session: Session, caplog: pytest.LogCaptureFixture
):
    pandas_df = pd.DataFrame(columns=["col1", "col2"])
    snowpark_schema = StructType(
        [
            StructField("col1", IntegerType(), nullable=False),
            StructField("col2", StringType(), nullable=False),
        ]
    )
    with caplog.at_level(level=logging.INFO, logger=LOGGER_NAME):
        result = generate_snowpark_dataframe(pandas_df, snowpark_schema, session)
    assert isinstance(result, DataFrame)
    assert result.count() == 0
    for substring in [str(snowpark_schema), str(len(pandas_df))]:
        assert substring in caplog.text


def test_generate_snowpark_dataframe_with_null_values(
    session: Session, caplog: pytest.LogCaptureFixture
):
    pandas_df = pd.DataFrame({"col1": [1, None], "col2": ["a", None]})
    snowpark_schema = StructType(
        [
            StructField("col1", IntegerType(), nullable=True),
            StructField("col2", StringType(), nullable=True),
        ]
    )
    with caplog.at_level(level=logging.INFO, logger=LOGGER_NAME):
        result = generate_snowpark_dataframe(pandas_df, snowpark_schema, session)
    assert isinstance(result, DataFrame)
    assert result.filter(result["col1"].isNull()).count() == 1
    assert result.filter(result["col2"].isNull()).count() == 1
    for substring in [str(snowpark_schema), str(len(pandas_df))]:
        assert substring in caplog.text


def test_temporary_random_seed_no_seed():
    initial_state = random.getstate()
    with temporary_random_seed():
        random_state_inside = random.getstate()
        assert random_state_inside != initial_state
    random_state_outside = random.getstate()
    assert random_state_outside == initial_state


def test_temporary_random_seed_with_seed():
    seed = 42
    initial_state = random.getstate()
    with temporary_random_seed(seed):
        random_state_inside = random.getstate()
        random.seed(seed)
        expected_state = random.getstate()
        assert random_state_inside == expected_state
    random_state_outside = random.getstate()
    assert random_state_outside == initial_state


def test_replace_surrogate_chars_no_surrogates():
    pandas_df = pd.DataFrame({"col1": ["a", "b", "c"], "col2": ["d", "e", "f"]})
    columns = ["col1", "col2"]
    result_df = replace_surrogate_chars(pandas_df, columns)
    assert result_df.equals(pandas_df)


def test_replace_surrogate_chars_with_surrogates():
    pandas_df = pd.DataFrame(
        {
            "col1": ["a", "\uD800", "c"],
            "col2": ["d", "\uD950", "f"],
            "col3": ["g", "\uDFFF", "i"],
        }
    )
    columns = ["col1", "col2", "col3"]
    expected_df = pd.DataFrame(
        {"col1": ["a", " ", "c"], "col2": ["d", " ", "f"], "col3": ["g", " ", "i"]}
    )
    result_df = replace_surrogate_chars(pandas_df, columns)
    assert result_df.equals(expected_df)


def test_replace_surrogate_chars_partial_columns():
    pandas_df = pd.DataFrame(
        {"col1": ["a", "\uD800", "c"], "col2": ["d", "\uDFFF", "f"]}
    )
    columns = ["col1"]
    expected_df = pd.DataFrame({"col1": ["a", " ", "c"], "col2": ["d", "\uDFFF", "f"]})
    result_df = replace_surrogate_chars(pandas_df, columns)
    assert result_df.equals(expected_df)


def test_replace_surrogate_chars_empty_dataframe():
    columns = ["col1", "col2"]
    pandas_df = pd.DataFrame(columns=columns)
    result_df = replace_surrogate_chars(pandas_df, columns)
    assert result_df.equals(pandas_df)


def test_replace_surrogate_chars_non_string_columns():
    pandas_df = pd.DataFrame(
        {"col1": [1, 2, 3], "col2": ["d", "\uDFFF", "f"], "col3": [None, np.NaN, "x"]}
    )
    columns = ["col1", "col2", "col3"]
    expected_df = pd.DataFrame(
        {"col1": [1, 2, 3], "col2": ["d", " ", "f"], "col3": [None, np.NaN, "x"]}
    )
    result_df = replace_surrogate_chars(pandas_df, columns)
    assert result_df.equals(expected_df)


def test_replace_surrogate_chars_missing_columns(caplog: pytest.LogCaptureFixture):
    pandas_df = pd.DataFrame(
        {
            "col1": ["a", "\uD800", "c"],
            "col2": ["d", "\uD950", "f"],
        }
    )
    columns = ["col1", "col2", "col3"]
    expected_df = pd.DataFrame({"col1": ["a", " ", "c"], "col2": ["d", " ", "f"]})

    with caplog.at_level(level=logging.WARNING, logger=LOGGER_NAME):
        result_df = replace_surrogate_chars(pandas_df, columns)
    assert result_df.equals(expected_df)
    assert "'col3' not found in DataFrame" in caplog.text



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

from functools import wraps
from typing import Callable


snowpark_strategies: dict[str, Callable] = {}
LOGGER = logging.getLogger(__name__)


def register_strategy(dtype: str) -> Callable:
    """Register a strategy for a given data type.

    Args:
        dtype: The data type to register the strategy for.

    Returns:
        The decorated function.

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        snowpark_strategies[dtype] = func
        LOGGER.debug("Registered '%s' function for '%s' type", func.__name__, dtype)
        return wrapper

    return decorator



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os

from pathlib import Path
from typing import Optional

from snowflake.hypothesis_snowpark.io_utils import EnvStrategy


class IODefaultStrategy(EnvStrategy):
    def mkdir(self, path: str, exist_ok: bool = False) -> None:
        os.makedirs(path, exist_ok=exist_ok)

    def folder_exists(self, path: str) -> bool:
        return os.path.isdir(path)

    def file_exists(self, path: str) -> bool:
        return os.path.isfile(path)

    def write(self, file_path: str, file_content: str, overwrite: bool = True) -> None:
        mode = "w" if overwrite else "x"
        with open(file_path, mode) as file:
            file.write(file_content)

    def read(
        self, file_path: str, mode: str = "r", encoding: Optional[str] = None
    ) -> str:
        with open(file_path, mode=mode, encoding=encoding) as file:
            return file.read()

    def read_bytes(self, file_path: str) -> bytes:
        with open(file_path, mode="rb") as f:
            return f.read()

    def ls(self, path: str, recursive: bool = False) -> list[str]:
        return glob.glob(path, recursive=recursive)

    def getcwd(self) -> str:
        return os.getcwd()

    def telemetry_path_files(self, path: str) -> Path:
        return Path(path)



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Optional

from snowflake.hypothesis_snowpark.io_utils import (
    EnvStrategy,
    IODefaultStrategy,
)
from snowflake.hypothesis_snowpark.singleton import Singleton


class IOFileManager(metaclass=Singleton):
    def __init__(self, strategy: Optional[EnvStrategy] = None):
        self.strategy = strategy or IODefaultStrategy()

    def mkdir(self, path: str, exist_ok: bool = False) -> None:
        return self.strategy.mkdir(path, exist_ok)

    def folder_exists(self, path: str) -> bool:
        return self.strategy.folder_exists(path)

    def file_exists(self, path: str) -> bool:
        return self.strategy.file_exists(path)

    def write(self, file_path: str, file_content: str, overwrite: bool = True) -> None:
        return self.strategy.write(file_path, file_content, overwrite)

    def read(
        self, file_path: str, mode: str = "r", encoding: Optional[str] = None
    ) -> str:
        return self.strategy.read(file_path, mode, encoding)

    def read_bytes(self, file_path: str) -> bytes:
        return self.strategy.read_bytes(file_path)

    def ls(self, path: str, recursive: bool = False) -> list[str]:
        return self.strategy.ls(path, recursive)

    def getcwd(self) -> str:
        return self.strategy.getcwd()

    def telemetry_path_files(self, path: str) -> Path:
        return self.strategy.telemetry_path_files(path)

    def set_strategy(self, strategy: EnvStrategy):
        """Set the strategy for file and directory operations.

        Args:
            strategy (EnvStrategy): The strategy to use for file and directory operations.

        """
        self.strategy = strategy


def get_io_file_manager():
    """Get the singleton instance of IOFileManager.

    Returns:
        IOFileManager: The singleton instance of IOFileManager.

    """
    return IOFileManager()



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re

from unittest.mock import MagicMock, Mock, patch

import pytest

from snowflake.hypothesis_snowpark import dataframe_strategy
from snowflake.hypothesis_snowpark.constants import CUSTOM_DATA_KEY, PANDERA_SCHEMA_KEY
from snowflake.hypothesis_snowpark.strategies import (
    _dataframe_strategy_from_json_schema,
    _dataframe_strategy_from_object_schema,
)
from snowflake.snowpark import Session


LOGGER_NAME = "snowflake.hypothesis_snowpark.strategies"


def test_dataframe_strategy_none_session(caplog: pytest.LogCaptureFixture):
    expected_error_msg = "Session cannot be None."
    with pytest.raises(ValueError, match=expected_error_msg), caplog.at_level(
        level=logging.ERROR, logger=LOGGER_NAME
    ):
        dataframe_strategy(schema="schema.json", session=None)
    assert expected_error_msg in caplog.text


def test_dataframe_strategy_invalid_schema(caplog: pytest.LogCaptureFixture):
    expected_error_msg = "Schema must be a path to a JSON schema file or a Pandera DataFrameSchema object."
    with pytest.raises(ValueError, match=expected_error_msg), caplog.at_level(
        level=logging.ERROR, logger=LOGGER_NAME
    ):
        dataframe_strategy(schema="123", session=Mock(spec=Session))
    assert expected_error_msg in caplog.text


def test_dataframe_strategy_invalid_json_file(caplog: pytest.LogCaptureFixture):
    mock_json_schema = {
        "key1": "value1",
        "key2": "value2",
    }
    expected_error_msg = "Invalid JSON schema. The JSON schema must contain 'pandera_schema' and 'custom_data' keys."

    with patch(
        "snowflake.hypothesis_snowpark.strategies.load_json_schema",
        return_value=mock_json_schema,
    ), pytest.raises(ValueError, match=expected_error_msg,), caplog.at_level(
        level=logging.ERROR, logger=LOGGER_NAME
    ):
        dataframe_strategy(
            schema="schema.json",
            session=Mock(spec=Session),
        )
    assert expected_error_msg in caplog.text


def test_dataframe_strategy_not_supported_dtypes(caplog: pytest.LogCaptureFixture):
    mock_json_schema = {
        "pandera_schema": {
            "columns": {
                "daytimeinterval_column": {"dtype": "timedelta64[ns]"},
                "map_column": {"dtype": "object"},
                "void_column": {"dtype": "object"},
                "struct_column": {"dtype": "object"},
            },
        },
        "custom_data": {
            "columns": [
                {"name": "daytimeinterval_column", "type": "daytimeinterval"},
                {"name": "map_column", "type": "map"},
                {"name": "void_column", "type": "void"},
                {"name": "struct_column", "type": "struct"},
            ]
        },
    }
    expected_error_msg = (
        "The following data types are not supported by the Snowpark DataFrame strategy: "
        "['daytimeinterval', 'map', 'void', 'struct']"
    )

    with patch(
        "snowflake.hypothesis_snowpark.strategies.load_json_schema",
        return_value=mock_json_schema,
    ), pytest.raises(ValueError, match=re.escape(expected_error_msg),), caplog.at_level(
        level=logging.ERROR, logger=LOGGER_NAME
    ):
        dataframe_strategy(
            schema="schema.json",
            session=Mock(spec=Session),
        )
    assert expected_error_msg in caplog.text


def test_object_schema_inner_dataframe_strategy_logs_exceptions(
    caplog: pytest.LogCaptureFixture,
):
    """Test that the inner _dataframe_strategy function logs exceptions when called with a Pandera schema object."""
    schema = MagicMock()
    exception_message = "Test exception"

    schema.strategy.side_effect = ValueError(exception_message)
    strategy = _dataframe_strategy_from_object_schema(
        schema=schema, session=MagicMock()
    )

    with pytest.raises(ValueError, match=exception_message), caplog.at_level(
        level=logging.ERROR, logger=LOGGER_NAME
    ):
        strategy.example()

    assert "An error occurred in _dataframe_strategy" in caplog.text
    assert exception_message in caplog.text


@patch("snowflake.hypothesis_snowpark.strategies.load_json_schema")
@patch("snowflake.hypothesis_snowpark.strategies.update_pandas_df_strategy")
def test_json_schema_inner_dataframe_strategy_logs_exceptions(
    mock_update_pandas_df_strategy: Mock,
    mock_load_json_schema: Mock,
    caplog: pytest.LogCaptureFixture,
):
    """Test that the inner _dataframe_strategy function logs exceptions when called with a JSON schema file."""
    json_schema_content = {
        PANDERA_SCHEMA_KEY: {"columns": {}},
        CUSTOM_DATA_KEY: {"columns": []},
    }
    exception_message = "Test exception"

    mock_load_json_schema.return_value = json_schema_content
    mock_update_pandas_df_strategy.side_effect = ValueError(exception_message)
    strategy = _dataframe_strategy_from_json_schema(
        schema="dummy.json", session=MagicMock()
    )

    with pytest.raises(ValueError, match=exception_message), caplog.at_level(
        level=logging.ERROR, logger=LOGGER_NAME
    ):
        strategy.example()

    assert "An error occurred in _dataframe_strategy" in caplog.text
    assert exception_message in caplog.text



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import shutil

from deepdiff import DeepDiff

from snowflake.hypothesis_snowpark.telemetry.telemetry import get_telemetry_manager
from snowflake.snowpark import Session

TEST_TELEMETRY_DATAFRAME_STRATEGIES_EXPECTED_DIRECTORY_NAME = (
    "test_telemetry_strategies_expected"
)

from snowflake.hypothesis_snowpark.telemetry.telemetry import TelemetryManager


def get_expected(file_name: str) -> str:
    current_directory_path = os.path.dirname(__file__)
    expected_file_path = os.path.join(
        current_directory_path,
        TEST_TELEMETRY_DATAFRAME_STRATEGIES_EXPECTED_DIRECTORY_NAME,
        file_name,
    )

    with open(expected_file_path) as f:
        return f.read().strip()


def remove_output_directory(telemetry_directory_path: str) -> None:
    if os.path.exists(telemetry_directory_path):
        shutil.rmtree(telemetry_directory_path)


def get_output_telemetry(telemetry_directory_path: str) -> str:
    files = os.listdir(telemetry_directory_path)
    files.sort(reverse=True)
    for file in files:
        if file.endswith(".json"):
            output_file_path = os.path.join(telemetry_directory_path, file)
            with open(output_file_path) as f:
                return f.read().strip()
    return "{}"


def validate_telemetry_file_output(
    telemetry_file_name: str, telemetry_directory_path: str
) -> None:
    telemetry_expected = get_expected(telemetry_file_name)
    telemetry_output = get_output_telemetry(telemetry_directory_path)

    telemetry_expected_obj = json.loads(telemetry_expected)
    telemetry_output_obj = json.loads(telemetry_output)
    exclude_telemetry_paths = [
        "root['timestamp']",
        "root['message']['metadata']['device_id']",
        "root['message']['metadata']",
        "root['message']['data']",
        "root['message']['driver_version']",
    ]

    diff_telemetry = DeepDiff(
        telemetry_expected_obj,
        telemetry_output_obj,
        ignore_order=True,
        exclude_paths=exclude_telemetry_paths,
    )
    remove_output_directory(telemetry_directory_path)
    get_telemetry_manager().sc_hypothesis_input_events = []

    assert diff_telemetry == {}
    assert isinstance(
        telemetry_output_obj.get("message")
        .get("metadata")
        .get("snowpark_checkpoints_version"),
        str,
    )


def reset_telemetry_util():
    connection = Session.builder.getOrCreate().connection
    connection._telemetry = TelemetryManager(
        connection._rest, connection.telemetry_enabled
    )
    return connection._telemetry



# Copyright 2025 Snowflake Inc.
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
import tempfile
from unittest.mock import patch
import pytest

from snowflake.hypothesis_snowpark.io_utils.io_file_manager import IODefaultStrategy
from snowflake.hypothesis_snowpark.io_utils.io_file_manager import (
    get_io_file_manager,
)
from snowflake.hypothesis_snowpark.singleton import Singleton


@pytest.fixture(scope="function")
def io_file_manager():
    Singleton._instances = {}
    return get_io_file_manager()


def test_default_mkdir(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_dir")

        # Act
        io_file_manager.mkdir(path)
        folder_exist = os.path.exists(path)

        # Assert
        assert folder_exist


def test_default_mkdir_fail(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_dir\0")

        # Act & Assert
        with pytest.raises(ValueError):
            io_file_manager.mkdir(path)


def test_default_folder_exists(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir)

        # Act
        result = io_file_manager.folder_exists(path)

        # Assert
        assert result


def test_default_folder_exists_fail(io_file_manager):
    # Act
    result_not_exist = io_file_manager.folder_exists("")

    # Assert
    assert result_not_exist is False


def test_default_folder_exists_fail_with_file(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_file.txt")
        with open(path, "w") as f:
            f.write("test")

        # Act
        result_not_folder = io_file_manager.folder_exists(path)

        # Assert
        assert result_not_folder is False


def test_default_file_exists(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_file.txt")
        with open(path, "w") as f:
            f.write("test")

        # Act
        result = io_file_manager.file_exists(path)

        # Assert
        assert result


def test_default_file_exists_fail(io_file_manager):
    # Act
    result_not_exists = io_file_manager.file_exists("")

    # Assert
    assert result_not_exists is False


def test_default_file_exists_fail_with_folder(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "temp_dir")
        os.makedirs(path)

        # Act
        result_not_file = io_file_manager.file_exists(path)

        # Assert
        assert result_not_file is False


def test_default_write(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_file.txt")
        content = "test"

        # Act
        io_file_manager.write(path, content)
        with open(path) as f:
            file_content = f.read()

        # Assert
        assert content == file_content


def test_default_write_fail(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        content = "test"

        # Act & Assert
        with pytest.raises(OSError):
            io_file_manager.write(temp_dir, content)


def test_default_read(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_file.txt")
        content = "test"
        with open(path, "w") as f:
            f.write(content)

        # Act
        result = io_file_manager.read(path)

        # Assert
        assert result == content


def test_default_read_fail(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_file.txt")

        # Act & Assert
        with pytest.raises(FileNotFoundError):
            io_file_manager.read(path)


def test_default_read_bytes(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_file.txt")
        content = "test"
        with open(path, "wb") as f:
            f.write(content.encode())

        # Act
        result = io_file_manager.read_bytes(path)

        # Assert
        assert result.decode() == content


def test_default_read_bytes_fail(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_file.txt")

        # Act & Assert
        with pytest.raises(FileNotFoundError):
            io_file_manager.read_bytes(path)


def test_default_ls(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_dir")
        os.makedirs(path)

        # Act
        result = io_file_manager.ls(temp_dir + "/*")

        # Assert
        assert len(result) == 1
        assert result[0] == path


def test_default_ls_case_2(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        file1 = os.path.join(temp_dir, "test.txt")
        file2 = os.path.join(temp_dir, "test2.py")

        with open(file1, "w") as f:
            f.write("test")

        with open(file2, "w") as f:
            f.write("test")

        # Act
        result = io_file_manager.ls(temp_dir + "/*.txt")

        # Assert
        assert len(result) == 1
        assert result[0] == file1


def test_default_ls_fail(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()):
        # Arrange
        path = None

        # Act
        result = io_file_manager.ls("")

        # Assert
        assert len(result) == 0


def test_default_getcwd(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()):
        # Act
        result = io_file_manager.getcwd()

        # Assert
        assert isinstance(result, str)


def test_default_telemetry_path_files(io_file_manager):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as temp_dir:
        # Arrange
        path = os.path.join(temp_dir, "test_dir")

        # Act
        result = io_file_manager.telemetry_path_files(path)

        # Assert
        assert result == Path(path)


def test_default_set_strategy(io_file_manager):
    try:
        # Arrange
        with patch(
            "snowflake.hypothesis_snowpark.io_utils.EnvStrategy"
        ) as MockStrategy:
            strategy = MockStrategy()

            # Act
            io_file_manager.set_strategy(strategy)

            # Assert
            assert io_file_manager.strategy == strategy
    finally:
        io_file_manager.set_strategy(IODefaultStrategy())


def test_default_get_io_file_manager_singleton():
    # Act
    result = get_io_file_manager()
    result2 = get_io_file_manager()

    # Assert
    assert result == result2
